==> blcr-0.5.0.tar.gz: blcr-0.5.0/cr_module/cr_compat.c <==
/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_compat.c,v 1.7 2007/02/01 00:48:19 phargrov Exp $
 *
 * This file is for calls to 64-bit kernels from 32-bit apps.
 */

#include <cr_module.h>

#if defined(CONFIG_COMPAT) /* File empty otherwise */

#include <asm/uaccess.h>

struct cr_compat_chkpt_req {
	compat_pid_t		cr_target;
	compat_int_t		cr_scope;	/* enum */
	compat_int_t		cr_fd;	
	compat_uint_t		cr_secs;
	compat_uptr_t		cr_opts;
	compat_int_t		dump_format;	/* enum */
	compat_int_t		signal;
	compat_uint_t		flags;
};

struct cr_compat_rstrt_args {
	compat_int_t			cr_fd;
	compat_uptr_t			cr_opts;
	compat_int_t			signal;
};

static inline int cr_compat_tv_call(int (*fn)(struct file *, void *),
				    struct file *file, unsigned long arg)
{
	struct compat_timeval __user *tvp = compat_ptr(arg);
	int retval = 0;

	if (tvp) {
		mm_segment_t oldfs;
		struct timeval tv;

		if (get_user(tv.tv_sec, &tvp->tv_sec) ||
		    get_user(tv.tv_usec, &tvp->tv_usec)) {
                        	return -EFAULT;
        	}

		oldfs = get_fs();
		set_fs(KERNEL_DS);
		retval = (*fn)(file, &tv);
		set_fs(oldfs);

		if (put_user(tv.tv_sec, &tvp->tv_sec) ||
		    put_user(tv.tv_usec, &tvp->tv_usec)) {
                        	return -EFAULT;
        	}
	} else {
		retval = (*fn)(file, NULL);
	}

	return retval;
}

static int cr_compat_chkpt_req(struct file *file, struct cr_compat_chkpt_req __user *req)
{
	mm_segment_t oldfs;
	struct cr_chkpt_req tmp;
	compat_uptr_t opts;
	int retval;

	if (get_user(tmp.cr_target, &req->cr_target) ||
	    get_user(tmp.cr_scope, &req->cr_scope) ||
	    get_user(tmp.cr_fd, &req->cr_fd) ||	
	    get_user(tmp.cr_secs, &req->cr_secs) ||
	    get_user(opts, &req->cr_opts) ||
	    get_user(tmp.dump_format, &req->dump_format) ||
	    get_user(tmp.signal, &req->signal) ||
	    get_user(tmp.flags, &req->flags) ||
	    get_user(opts, &req->cr_fd)) {
		return -EFAULT;
	}
	tmp.cr_opts = compat_ptr(opts);

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = cr_chkpt_req(file, &tmp);
	set_fs(oldfs);

	return retval;
}

static int cr_compat_rstrt_request_restart(struct file *file, struct cr_compat_rstrt_args __user *req)
{
	mm_segment_t oldfs;
	struct cr_rstrt_args tmp;
	compat_uptr_t opts;
	int retval;

	if (get_user(tmp.cr_fd, &req->cr_fd) ||
	    get_user(opts, &req->cr_opts)) {
		return -EFAULT;
	}
	tmp.cr_opts = compat_ptr(opts);

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = cr_rstrt_request_restart(file, &tmp);
	set_fs(oldfs);

	return retval;
}

/* Wrapper for 32bit callers */
long cr_compat_ctrl_ioctl(struct file *file, unsigned op, unsigned long arg)
{
	CR_KTRACE_FUNC_ENTRY("op=%08x arg=0x%lx", op, arg);

	switch(CR_IOCTL32_UNMAP(op)) {
	//
	// Calls from libcr:
	//
	case CR_OP_HAND_CHKPT:
		return cr_dump_self(file, arg);

	case CR_OP_HAND_ABORT:
		return cr_abort_checkpoint(file, arg);

	case CR_OP_HAND_CONT:
		return cr_cont_complete(file);

	case CR_OP_HAND_RSTRT:
		return cr_rstrt_complete(file);

	case CR_OP_HAND_SUSP:
		return cr_compat_tv_call(cr_suspend, file, arg);

	case CR_OP_HAND_PHASE1:
		return cr_phase1_register(file, compat_ptr(arg));

	case CR_OP_HAND_PHASE2:
		return cr_phase2_register(file, compat_ptr(arg));

	case CR_OP_HAND_DEST:
		return cr_chkpt_dest(file, compat_ptr(arg));

	case CR_OP_HAND_SRC:
		return cr_rstrt_src(file, compat_ptr(arg));

	//
	// Calls from cr_checkpoint:
	//
	case CR_OP_CHKPT_REQ:
		return cr_compat_chkpt_req(file, compat_ptr(arg));

	case CR_OP_CHKPT_DONE:
		return cr_compat_tv_call(cr_chkpt_done, file, arg);

	case CR_OP_CHKPT_REAP:
		return cr_chkpt_reap(file);

	//
	// Calls from cr_restart:
	//
	case CR_OP_RSTRT_REQ:
		return cr_compat_rstrt_request_restart(file, compat_ptr(arg));

	case CR_OP_RSTRT_DONE:
		return cr_compat_tv_call(&cr_rstrt_done, file, arg);

	case CR_OP_RSTRT_REAP:
		return cr_rstrt_reap(file);

	case CR_OP_RSTRT_CHILD:
		return cr_rstrt_child(file);

	case CR_OP_RSTRT_PROCS:
		return cr_rstrt_procs(file, compat_ptr(arg));

	//
	// General calls
	//
	case CR_OP_VERSION:
	    {
		// Simple enough to do here
		// User-space must have equal MAJOR and MINOR <= kernel
		// If we supported multiple major versions then this is
		// where we could do the "personallity switching".
		unsigned long major = (arg >> 16);
		unsigned long minor = (arg & 0xffff);
                // If major==0, must match minor; otherwise user-MINOR <= kernel-MINOR
                int match = (major == CR_MODULE_MAJOR) &&
                            (major ? (minor <= CR_MODULE_MINOR)
                                   : (minor == CR_MODULE_MINOR));
		if (!match) {
		    CR_WARN("request from pid %d for unsupported version %d.%d",
			    (int)current->pid, (int)major, (int)minor);
		}
		return match ? 0 : -CR_EVERSION;
	    }

	default:
		CR_KTRACE_BADPARM("unknown op %x", _IOC_NR(op));
		return -ENOTTY;
	}
}

#endif /* CONFIG_COMPAT */

==> blcr-0.6.2.tar.gz: blcr-0.6.2/cr_module/cr_compat.c <==
/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_compat.c,v 1.10 2007/07/17 22:27:28 eroman Exp $
 *
 * This file is for calls to 64-bit kernels from 32-bit apps.
 */

#include <cr_module.h>

#if defined(CONFIG_COMPAT) /* File empty otherwise */

#include <asm/uaccess.h>

struct cr_compat_chkpt_req {
	compat_pid_t		cr_target;
	compat_int_t		cr_scope;	/* enum */
	compat_int_t		cr_fd;	
	compat_uint_t		cr_secs;
	compat_uptr_t		cr_opts;
	compat_int_t		dump_format;	/* enum */
	compat_int_t		signal;
	compat_uint_t		flags;
};

struct cr_compat_chkpt_fwd {
	compat_pid_t		cr_target;
	compat_int_t		cr_scope;	/* enum */
};

struct cr_compat_rstrt_args {
	compat_int_t			cr_fd;
	compat_uptr_t			cr_opts;
	compat_int_t			signal;
};

struct cr_compat_chkpt_info {
	compat_pid_t		requester;
	compat_pid_t		target;
	compat_int_t		scope;
	compat_int_t		signal;
	compat_uptr_t		dest;
};

static inline int cr_compat_tv_call(int (*fn)(struct file *, void *),
				    struct file *file, unsigned long arg)
{
	struct compat_timeval __user *tvp = compat_ptr(arg);
	int retval = 0;

	if (tvp) {
		mm_segment_t oldfs;
		struct timeval tv;

		if (get_user(tv.tv_sec, &tvp->tv_sec) ||
		    get_user(tv.tv_usec, &tvp->tv_usec)) {
                        	return -EFAULT;
        	}

		oldfs = get_fs();
		set_fs(KERNEL_DS);
		retval = (*fn)(file, &tv);
		set_fs(oldfs);

		if (put_user(tv.tv_sec, &tvp->tv_sec) ||
		    put_user(tv.tv_usec, &tvp->tv_usec)) {
                        	return -EFAULT;
        	}
	} else {
		retval = (*fn)(file, NULL);
	}

	return retval;
}

static int cr_compat_chkpt_req(struct file *file, struct cr_compat_chkpt_req __user *req)
{
	mm_segment_t oldfs;
	struct cr_chkpt_req tmp;
	compat_uptr_t opts;
	int retval;

	if (get_user(tmp.cr_target, &req->cr_target) ||
	    get_user(tmp.cr_scope, &req->cr_scope) ||
	    get_user(tmp.cr_fd, &req->cr_fd) ||	
	    get_user(tmp.cr_secs, &req->cr_secs) ||
	    get_user(opts, &req->cr_opts) ||
	    get_user(tmp.dump_format, &req->dump_format) ||
	    get_user(tmp.signal, &req->signal) ||
	    get_user(tmp.flags, &req->flags) ||
	    get_user(opts, &req->cr_fd)) {
		return -EFAULT;
	}
	tmp.cr_opts = compat_ptr(opts);

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = cr_chkpt_req(file, &tmp);
	set_fs(oldfs);

	return retval;
}

static int cr_compat_chkpt_fwd(struct file *file, struct cr_compat_chkpt_fwd __user *req)
{
	mm_segment_t oldfs;
	struct cr_fwd_req tmp;
	int retval;

	if (get_user(tmp.cr_target, &req->cr_target) ||
	    get_user(tmp.cr_scope, &req->cr_scope)) {
		return -EFAULT;
	}

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = cr_chkpt_fwd(file, &tmp);
	set_fs(oldfs);

	return retval;
}

static int cr_compat_rstrt_request_restart(struct file *file, struct cr_compat_rstrt_args __user *req)
{
	mm_segment_t oldfs;
	struct cr_rstrt_args tmp;
	compat_uptr_t opts;
	int retval;

	if (get_user(tmp.cr_fd, &req->cr_fd) ||
	    get_user(opts, &req->cr_opts)) {
		return -EFAULT;
	}
	tmp.cr_opts = compat_ptr(opts);

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = cr_rstrt_request_restart(file, &tmp);
	set_fs(oldfs);

	return retval;
}

static int cr_compat_chkpt_info(struct file *file, struct cr_compat_chkpt_info __user *info)
{
	mm_segment_t oldfs;
	struct cr_chkpt_info tmp;
	compat_uptr_t dest;
	int retval;

	if (get_user(dest, &info->dest)) {
		return -EFAULT;
	}
	tmp.dest = compat_ptr(dest);

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = cr_chkpt_info(file, &tmp);
	set_fs(oldfs);

	if (!retval &&
	    (put_user((compat_pid_t)tmp.requester, &info->requester) ||
	     put_user((compat_pid_t)tmp.target, &info->target) ||
	     put_user((compat_pid_t)tmp.scope, &info->scope) ||
	     put_user((compat_int_t)tmp.signal, &info->signal))) {
		retval = -EFAULT;
	}

	return retval;
}

/* Wrapper for 32bit callers */
long cr_compat_ctrl_ioctl(struct file *file, unsigned op, unsigned long arg)
{
	CR_KTRACE_FUNC_ENTRY("op=%08x arg=0x%lx", op, arg);

	switch(CR_IOCTL32_UNMAP(op)) {
	//
	// Calls from libcr:
	//
	case CR_OP_HAND_CHKPT:
		return cr_dump_self(file, arg);

	case CR_OP_HAND_ABORT:
		return cr_abort_checkpoint(file, arg);

	case CR_OP_HAND_CONT:
		return cr_cont_complete(file);

	case CR_OP_HAND_RSTRT:
		return cr_rstrt_complete(file);

	case CR_OP_HAND_SUSP:
		return cr_compat_tv_call(cr_suspend, file, arg);

	case CR_OP_HAND_PHASE1:
		return cr_phase1_register(file, compat_ptr(arg));

	case CR_OP_HAND_PHASE2:
		return cr_phase2_register(file, compat_ptr(arg));

	case CR_OP_HAND_SRC:
		return cr_rstrt_src(file, compat_ptr(arg));

	case CR_OP_HAND_CHKPT_INFO:
		return cr_compat_chkpt_info(file, compat_ptr(arg));

	//
	// Calls from cr_checkpoint:
	//
	case CR_OP_CHKPT_REQ:
		return cr_compat_chkpt_req(file, compat_ptr(arg));

	case CR_OP_CHKPT_DONE:
		return cr_compat_tv_call(cr_chkpt_done, file, arg);

	case CR_OP_CHKPT_REAP:
		return cr_chkpt_reap(file);

	case CR_OP_CHKPT_FWD:
		return cr_compat_chkpt_fwd(file, compat_ptr(arg));

	//
	// Calls from cr_restart:
	//
	case CR_OP_RSTRT_REQ:
		return cr_compat_rstrt_request_restart(file, compat_ptr(arg));

	case CR_OP_RSTRT_DONE:
		return cr_compat_tv_call(&cr_rstrt_done, file, arg);

	case CR_OP_RSTRT_REAP:
		return cr_rstrt_reap(file);

	case CR_OP_RSTRT_CHILD:
		return cr_rstrt_child(file);

	case CR_OP_RSTRT_PROCS:
		return cr_rstrt_procs(file, compat_ptr(arg));

	//
	// General calls
	//
	case CR_OP_VERSION:
	    {
		// Simple enough to do here
		// User-space must have equal MAJOR and MINOR <= kernel
		// If we supported multiple major versions then this is
		// where we could do the "personallity switching".
		unsigned long major = (arg >> 16);
		unsigned long minor = (arg & 0xffff);
                // If major==0, must match minor; otherwise user-MINOR <= kernel-MINOR
                int match = (major == CR_MODULE_MAJOR) &&
                            (major ? (minor <= CR_MODULE_MINOR)
                                   : (minor == CR_MODULE_MINOR));
		if (!match) {
		    CR_WARN("request from pid %d for unsupported version %d.%d",
			    (int)current->pid, (int)major, (int)minor);
		}
		return match ? 0 : -CR_EVERSION;
	    }

	default:
		CR_KTRACE_BADPARM("unknown op %x", _IOC_NR(op));
		return -ENOTTY;
	}
}

#endif /* CONFIG_COMPAT */
